/******************************************************************************
* Module	:	Python Trie Module.
*
* Author	:	John W. M. Stevens
*
* License	:	This library is free software; you can redistribute it
*				and/or modify it under the terms of the GNU Library
*				General Public License as published by the Free Software
*				Foundation; either version 2 of the License, or (at your
*				option) any later version.
*
*				This library is distributed in the hope that it will be
*				useful, but WITHOUT ANY WARRANTY; without even the implied
*				warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
*				PURPOSE.  See the GNU Library General Public License for
*				more details.
*
*				You should have received a copy of the GNU Library General
*				Public License along with this library; if not, write to
*				the Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
*				MA 02139, USA.
******************************************************************************/

#include    <stdio.h>
#include    <stdlib.h>
#include    <string.h>
#include    <Python.h>

#include	"TriePy.h"

/*	Python 'type' structure.	*/
PyTypeObject    TrieObjectType;

/*	Type testing macro.	*/
#define	is_TrieObject(v)       ((v)->ob_type == &TrieObjectType)

/*	Error return object.
*
*	Note: Failing to generate this on the fly may, or may not violate
*	multi-threading restrictions. . .
*/
PyObject        *trieErrorObject;

/*  Error reporting macro.  */
#define TrieError(s)  \
	{ PyErr_SetString(trieErrorObject, (s)); return NULL; }

#define	SetError(s)	\
	PyErr_SetString(trieErrorObject, (s))

/*=============================================================================
||	Private Functions.	Built as pure C functions so that they can be
||	easily used in Python, Objective C and C++ with only minor modifications
||	for generic type, list objects, etc..
=============================================================================*/

/*-----------------------------------------------------------------------------
| Method	:	copy --- Copy a trie.
-----------------------------------------------------------------------------*/

static
TrieNode	*
copy(
	TrieNode	*node)
{
	register	int			i;
	auto		TrieNode	*newNode;
	auto		int			size;

	/*	Create a new node of the proper size.	*/
	size = sizeof( TrieNode ) + node->used * sizeof( TrieElement );
	newNode = calloc(1, size);

	/*	Copy.	*/
	memcpy(
		newNode,
		node,
		sizeof( TrieNode ) + node->used * sizeof( TrieElement ));

	/*	Copy children.	*/
	for (i = 0; i < newNode->used; i++)
	{
		/*	Data reference count increment.	*/
		if ( newNode->els[i].data )
			Attach( newNode->els[i].data );

		/*	Make copy of child node, modify pointer.	*/
		if ( newNode->els[i].child )
			newNode->els[i].child = copy( newNode->els[i].child );
	}

	/*	Return new node.	*/
	return newNode;
}

/*-----------------------------------------------------------------------------
| Method	:	search --- Do binary search within a node.
-----------------------------------------------------------------------------*/

static
int
search(
	TrieNode	*node,
	char		c,
	int			*lo,
	int			*mid,
	int			*hi)
{
	auto	int		ret;

	/*	Do binary search.	*/
	*hi = node->used - 1;
	*lo = 0;
	do
	{
		/*	Calculate mid point, do comparison.	*/
		*mid = (*hi + *lo) >> 1;
		ret  = c - node->els[*mid].c;

		/*	Adjust search range.	*/
		if (ret <= 0)
			*hi = *mid - 1;
		if (ret >= 0)
			*lo = *mid + 1;
	} while (*hi >= *lo);

	/*	Return search results.	*/
	return ret;
}

/*-----------------------------------------------------------------------------
| Method	:	insert --- Insert a node.
-----------------------------------------------------------------------------*/

static
TrieNode	*
insert(
	TrieNode	*node,
	char		*str,
	unsigned	*total,
	int			*count,
	GenObject	data)
{
	auto		TrieNode	*tmp;
	auto		TrieNode	*child;
	auto		unsigned	size;
	auto		int			hi;
	auto		int			mid;
	auto		int			lo;
	auto		int			ret;

	/*	If node is NULL, allocate one.	*/
	if (node == NULL)
	{
		/*	Allocate a TrieNode.	*/
		tmp = calloc(1, sizeof( TrieNode ));

		/*	Insert character.	*/
		mid = 0;
		tmp->used = 1;
		tmp->els[mid].c        = *str;
		tmp->els[mid].subCount = 0;
		tmp->els[mid].child    = child = NULL;
		tmp->els[mid].data     = GenNULL;
	}
	else if ((ret = search(node, *str, &lo, &mid, &hi)) != 0)
	{
		/*	Expand and add to node.	*/
		size = sizeof( TrieNode ) + (node->used + 1) * sizeof( TrieElement );
		tmp = calloc(1, size);

		/*	Copy.	*/
		memcpy(
			tmp,
			node,
			sizeof( TrieNode ) + node->used * sizeof( TrieElement ));

		/*	Make space.	*/
		mid = lo;
		if (lo < tmp->used)
			memmove(
				tmp->els + mid + 1,
				tmp->els + mid,
				(tmp->used - mid) * sizeof( TrieElement ));

		/*	Insert character.	*/
		tmp->els[mid].c        = *str;
		tmp->els[mid].subCount = 0;
		tmp->els[mid].child    = child = NULL;
		tmp->els[mid].data     = GenNULL;
		tmp->used++;
	}
	else
	{
		child = node->els[mid].child;
		tmp   = node;
	}

	/*	If not at end of string, continue insert.	*/
	if ( str[1] )
	{
		/*	Recurse.	*/
		tmp->els[mid].child =
			insert(
				tmp->els[mid].child,
				str + 1,
				total,
				count,
				data);

		/*	Increase sub count.	*/
		tmp->els[mid].subCount += *count;
	}
	else
	{
		/*	Decrement reference count on object to be remapped.	*/
		if ( tmp->els[mid].data )
			Dispose( tmp->els[mid].data );
		else
		{
			(*total)++;
			(*count) = 1;
			tmp->els[mid].subCount += *count;
		}

		/*	Increment reference count, save mapped object.	*/
		Attach( data );
		tmp->els[mid].data  = data;
	}

	/*	Return node.	*/
	return tmp;
}

/*-----------------------------------------------------------------------------
| Method	:	remove --- Remove a trie string.
-----------------------------------------------------------------------------*/

static
GenObject
removeNode(
	TrieNode	*node,
	char		*str,
	int			*count,
	unsigned	*total)
{
	auto	GenObject	data;
	auto	int			hi;
	auto	int			mid;
	auto	int			lo;
	auto	int			ret;

	/*	If found, check for continuation.	*/
	if ((ret = search(node, *str, &lo, &mid, &hi)) != 0)
		return GenNULL;

	/*	If not end of string, and not end of trie branch, continue.	*/
	if (str[1] && node->els[mid].child != NULL)
	{
		/*	Go recurse.	*/
		data = removeNode(node->els[mid].child, str + 1, count, total);

		/*	Check for empty.	*/
		if (node->els[mid].child->used < 1)
		{
			free( node->els[mid].child );
			node->els[mid].child = NULL;
		}
	}
	else if ( str[1] )
		return GenNULL;
	else
	{
		/*	Get data reference, decrement count.	*/
		if ((data = node->els[mid].data) != NULL)
		{
			*count = -1;
			(*total)--;
		}

		/*	Zero out data.	*/
		node->els[mid].subCount += *count;
		node->els[mid].data = GenNULL;
		Dispose( data );
	}

	/*	Do we need to remove this?	*/
	if (node->els[mid].child == NULL &&
		node->els[mid].data  == GenNULL)
	{
		/*	Decrement nodes used.	*/
		node->used--;

		/*	Adjust.	*/
		if (mid < node->used)
			memcpy(
				node->els + mid,
				node->els + mid + 1,
				sizeof( TrieElement ) * (node->used - mid));
	}

	/*	Return the data object reference.	*/
	return data;
}

/*-----------------------------------------------------------------------------
| Method	:	traverse --- Traverse the Trie, performing the necesary
|				operation.
-----------------------------------------------------------------------------*/

static
void
doOperation(
	TrieNode	*node,
	int			i,
	Traversal	*trav,
	int			index)
{
	/*	If this character has 'data', then append to list.	*/
	switch ( trav->operation )
	{
	case TRIE_DATA:
		ListAppend(trav->list, node->els[i].data);
		break;
	case TRIE_DEBUG:
		{
			unsigned		j;
			
			for (j = 0; j < index; j++)
				fputs("  ", stdout);
			printf("%c: (%d:%d) ",
				node->els[i].c,
				node->els[i].subCount,
				node->els[i].data->ob_refcnt);
			PrintObject(node->els[i].data, stdout, 0);
			fputs("\n", stdout);
		}
		break;
	case TRIE_FREE:
		Py_XDECREF( node->els[i].data );
		break;
	case TRIE_INTERNAL:
		trav->doFunc(
			trav->bfr,
			node->els[i].data,
			trav->data);
		break;
	case TRIE_KEYS:
		ListAppend(trav->list,
				   MakeStringObjectFromCString( trav->bfr ));
		break;
	case TRIE_PERFORM:
		{
			/*	Perform action.	*/
			PerformObject(
				trav->list,
				trav->bfr,
				node->els[i].data)
		}
		break;
	case TRIE_PRINT:
		/*	Check for count.	*/
		if ( trav->count )
			fputs(", ", trav->fp);

		/*	Print.	*/
		fprintf(trav->fp, "'%s': ", trav->bfr);
		PrintObject(node->els[i].data, trav->fp, 0);
		trav->count++;
		break;
	}
}

static
void
traverse(
	TrieNode	*node,
	Traversal	*trav,
	int			index)
{
	register	unsigned		i;

	/*	Traverse node from low to high.	*/
	for (i = 0; i < node->used; i++)
	{
		/*	Append to buffer.	*/
		if (index >= trav->len)
		{
			trav->len += 16;
			trav->bfr  = realloc(trav->bfr, trav->len);
		}

		/*	Put character in buffer.	*/
		trav->bfr[index]     = node->els[i].c;
		trav->bfr[index + 1] = '\0';

		/*	Do operation.	*/
		if (node->els[i].data != GenNULL)
			doOperation(node, i, trav, index);

		/*	Go down.	*/
		if ( node->els[i].child )
			traverse(
				node->els[i].child,
				trav,
				index + 1);
	}
}

/*-----------------------------------------------------------------------------
| Method	:	traverseReverse --- Traverse the Trie, performing the necesary
|				operation.
-----------------------------------------------------------------------------*/

static
void
traverseReverse(
	TrieNode	*node,
	Traversal	*trav,
	int			index)
{
	register	int		i;

	/*	Traverse node from high to low.	*/
	for (i = node->used - 1; i >= 0; i--)
	{
		/*	Append to buffer.	*/
		if (index >= trav->len)
		{
			trav->len += 16;
			trav->bfr  = realloc(trav->bfr, trav->len);
		}

		/*	Put character in buffer.	*/
		trav->bfr[index]     = node->els[i].c;
		trav->bfr[index + 1] = '\0';

		/*	If this character has 'data', then append to list.	*/
		if (node->els[i].data != GenNULL)
			doOperation(node, i, trav, index);
		else if (trav->operation == TRIE_DEBUG)
		{
			unsigned		j;
			
			for (j = 0; j < index; j++)
				fputs("  ", stdout);
			printf("%c: (%d:NA)\n",
				node->els[i].c,
				node->els[i].subCount);
		}

		/*	Go down.	*/
		if ( node->els[i].child )
			traverseReverse(
				node->els[i].child,
				trav,
				index + 1);
	}
}

/*-----------------------------------------------------------------------------
| Method	:	find --- Find matching string in Trie.
-----------------------------------------------------------------------------*/

static
GenObject
find(
	TrieNode	*node,
	char		*str)
{
	auto		TrieNode	*child;
	auto		int			hi;
	auto		int			mid;
	auto		int			lo;
	auto		int			ret;
	auto		GenObject	returnObject;

	/*	If not found, return.	*/
	if ((ret = search(node, *str, &lo, &mid, &hi)) != 0)
		return GenNULL;

	/*	Get pointer to child.	*/
	child = node->els[mid].child;

	/*	If not at end of string, continue search.	*/
	if (child && str[1])
		if ((returnObject =
			find(
				child,
				str + 1)) == GenNULL && node->els[ mid ].data)
			return node->els[ mid ].data;
		else
			return returnObject;

	/*	Return.	*/
	return node->els[mid].data;
}

/*-----------------------------------------------------------------------------
| Method	:	freeTrie --- Free the trie, completely.
-----------------------------------------------------------------------------*/

static
void
freeTrie(
	TrieNode	*node)
{
	register	int			i;

	/*	Do depth first traversal.	*/
	for (i = 0; i < node->used; i++)
	{
		/*	Dispose of each object we run across.	*/
		if ( node->els[i].data )
			Dispose( node->els[i].data );

		/*	Process child, if there is one.	*/
		if ( node->els[i].child )
			freeTrie( node->els[i].child );
	}

	/*	Free this node, return.	*/
	free( node );
}

/*-----------------------------------------------------------------------------
| Method	:	traversePrefix --- Traverse the Trie to the matching
|				prefix point, then traverse the sub-trie as appropriate.
-----------------------------------------------------------------------------*/

static
void
traversePrefix(
	TrieNode	*node,
	char		*str,
	Traversal	*trav,
	int			index)
{
	auto		TrieNode	*child;
	auto		int			hi;
	auto		int			mid;
	auto		int			lo;
	auto		int			ret;

	/*	If not found, return.	*/
	if ((ret = search(node, str[index], &lo, &mid, &hi)) != 0)
		return;

	/*	Get pointer to child.	*/
	child = node->els[mid].child;

	/*	If not at end of string, continue search.	*/
	if ( str[index + 1] )
	{
		/*	If the prefix does not exist in the Trie, simply return.	*/
		if ( child )
			return traversePrefix(child, str, trav, index + 1);
		else
			return;
	}

	/*	Build bfr value.	*/
	trav->bfr = strdup( str );
	trav->len = index + 1;

	/*	Perform operation.	*/
	if (node->els[mid].data != GenNULL)
		doOperation(node, mid, trav, index);

	/*	Traverse sub-trie.	*/
	if ( child )
		traverse(child, trav, index + 1);
}

/*=============================================================================
||	Extra stuff specific for Tries.
=============================================================================*/

static
PyObject	*
complete(
	PyObject	*self,
	PyObject	*args)
{
	auto	GenObject		call;
	auto	GenObject		list;
	auto	TrieObject		*trie = (TrieObject *) self;
	auto	char			*str;
	auto	Traversal		trav;

	/*	Build a new list.	*/
	list = PyList_New( 0 );

    /*  Parse the arguments.    */
    if (! PyArg_ParseTuple(args,
                           "s",
						   &str))
        return NULL;

	/*	Run routine to traverse entire trie, returning sorted
	*	list of keys.
	*/
	if ( trie->total )
	{
		/*	Build traversal control block.	*/
		trav.operation = TRIE_KEYS;
		trav.bfr       = NULL;
		trav.len       = 0;
		trav.count     = 0;
		trav.list      = list;

		/*	Do recursive traversal.	*/
		traversePrefix(
			trie->root,
			str,
			&trav,
			0);

		/*	Free buffer.	*/
		if ( trav.bfr )
			free( trav.bfr );
	}

	/*	Return completion list.	*/
	return list;
}

static
PyObject	*
data(
	PyObject	*self,
	PyObject	*args)
{
	auto	GenObject		list;
	auto	TrieObject		*trie = (TrieObject *) self;
	auto	Traversal		trav;

	/*	Build a new list.	*/
	list = PyList_New( 0 );

	/*	Run routine to traverse entire trie, returning
	*	list of data sorted by key value.
	*/
	if ( trie->total )
	{
		/*	Build traversal control block.	*/
		trav.operation = TRIE_DATA;
		trav.bfr       = NULL;
		trav.len       = 0;
		trav.count     = 0;
		trav.list      = list;

		/*	Do recursive traversal.	*/
		traverse(
			trie->root,
			&trav,
			0);

		/*	Free buffer.	*/
		if ( trav.bfr )
			free( trav.bfr );
	}

	/*	Increment reference count, return object.	*/
	return list;
}

static
PyObject	*
debug(
	PyObject	*self,
	PyObject	*args)
{
	auto	GenObject		list;
	auto	TrieObject		*trie = (TrieObject *) self;
	auto	Traversal		trav;

	/*	Run routine to traverse entire trie, returning
	*	list of data sorted by key value.
	*/
	printf("Reference Count: %d\n", trie->ob_refcnt);
	printf("Total          : %u\n", trie->total);
	if ( trie->total )
	{
		/*	Build traversal control block.	*/
		trav.operation = TRIE_DEBUG;
		trav.bfr       = NULL;
		trav.len       = 0;
		trav.count     = 0;

		/*	Do recursive traversal.	*/
		traverseReverse(
			trie->root,
			&trav,
			0);

		/*	Free buffer.	*/
		if ( trav.bfr )
			free( trav.bfr );
	}

	/*	Increment reference count, return object.	*/
	Py_XINCREF( Py_None );
	return Py_None;
}

static
PyObject	*
has_key(
	PyObject	*self,
	PyObject	*args)
{
	auto	char		*cp;
	auto	PyObject	*retObj;
	auto	TrieObject	*trie = (TrieObject *) self;
	auto	char		*keyString;

    /*  Parse the arguments.    */
    if (! PyArg_ParseTuple(args,
                           "s",
                           &keyString))
        return NULL;

	/*	Find the string in the trie.	*/
	if ((retObj = find(trie->root, keyString)) == NULL)
		return Py_BuildValue("i", 0);

	/*	Increment reference count, return object.	*/
	return Py_BuildValue("i", 1);
}

static
PyObject	*
keys(
	PyObject	*self,
	PyObject	*args)
{
	auto	GenObject		list;
	auto	TrieObject		*trie = (TrieObject *) self;
	auto	Traversal		trav;

	/*	Build a new list.	*/
	list = PyList_New( 0 );

	/*	Run routine to traverse entire trie, returning sorted
	*	list of keys.
	*/
	if ( trie->total )
	{
		/*	Build traversal control block.	*/
		trav.operation = TRIE_KEYS;
		trav.bfr       = NULL;
		trav.len       = 0;
		trav.count     = 0;
		trav.list      = list;

		/*	Do recursive traversal.	*/
		traverse(
			trie->root,
			&trav,
			0);

		/*	Free buffer.	*/
		if ( trav.bfr )
			free( trav.bfr );
	}

	/*	Increment reference count, return object.	*/
	return list;
}

static
PyObject	*
perform(
	PyObject	*self,
	PyObject	*args)
{
	auto	GenObject		call;
	auto	TrieObject		*trie = (TrieObject *) self;
	auto	Traversal		trav;

    /*  Parse the arguments.    */
    if (! PyArg_ParseTuple(args,
                           "O",
                           &call))
        return NULL;

	/*	Run routine to traverse entire trie, returning sorted
	*	list of keys.
	*/
	if ( trie->total )
	{
		/*	Build traversal control block.	*/
		trav.operation = TRIE_PERFORM;
		trav.bfr       = NULL;
		trav.len       = 0;
		trav.count     = 0;
		trav.list      = call;

		/*	Do recursive traversal.	*/
		traverse(
			trie->root,
			&trav,
			0);

		/*	Free buffer.	*/
		if ( trav.bfr )
			free( trav.bfr );
	}

	/*	Increment reference count, return object.	*/
	Py_XINCREF( Py_None );
	return Py_None;
}

static
PyObject	*
prefix(
	PyObject	*self,
	PyObject	*args)
{
	auto	GenObject		call;
	auto	TrieObject		*trie = (TrieObject *) self;
	auto	char			*str;
	auto	Traversal		trav;

    /*  Parse the arguments.    */
    if (! PyArg_ParseTuple(args,
                           "sO",
						   &str,
                           &call))
        return NULL;

	/*	Run routine to traverse entire trie, returning sorted
	*	list of keys.
	*/
	if ( trie->total )
	{
		/*	Build traversal control block.	*/
		trav.operation = TRIE_PERFORM;
		trav.bfr       = NULL;
		trav.len       = 0;
		trav.count     = 0;
		trav.list      = call;

		/*	Do recursive traversal.	*/
		traversePrefix(
			trie->root,
			str,
			&trav,
			0);

		/*	Free buffer.	*/
		if ( trav.bfr )
			free( trav.bfr );
	}

	/*	Increment reference count, return object.	*/
	Py_XINCREF( Py_None );
	return Py_None;
}

static  PyMethodDef     TrieObject_methods[] = {
	{	"complete",		complete,		1	},
	{	"debug",		debug,			1	},
	{	"data",			data,			1	},
	{	"has_key",		has_key,		1	},
	{	"keys",			keys,			1	},
	{	"perform",		perform,		1	},
	{	"prefix",		prefix,			1	},
	{	NULL,			NULL	}
};

/*=============================================================================
||	Life Cycle Methods and Conformant Implementation of PyObject Protocol.
=============================================================================*/

static
PyObject    *
Trie(
    PyObject    *self,
    PyObject    *args)
{
    auto	TrieObject	*trieObj;

    /*  Get new push button object. */
    trieObj = PyObject_NEW(
        TrieObject,
        &TrieObjectType);

	/*	Set to be empty.	*/
	trieObj->root  = NULL;
	trieObj->total = 0u;

    /*  Return the new object.  */
    return (PyObject *) trieObj;
}

static
PyObject	*
getattr(PyObject	*self,
		char		*name)
{
	/*	Return a list of strings if asked for the members.	*/
	if (strcmp(name, "__members__") == 0)
		return Py_BuildValue("[s]",
					"debug",
					"data",
					"keys",
					"perform",
					"prefix");
	return Py_FindMethod(TrieObject_methods, self, name);
}

static
int
print(
	PyObject	*self,
	FILE		*fp,
	int			flags)
{
	register	int				i;
	auto		TrieObject		*trie = (TrieObject *) self;
	auto		Traversal		trav;

	/*	Shut the compiler up.	*/
	flags;

	/*	Print contents.	*/
	fputs("{", fp);
	if ( trie->root )
	{
		/*	Build traversal control block.	*/
		trav.operation = TRIE_PRINT;
		trav.bfr       = NULL;
		trav.len       = 0;
		trav.count     = 0;
		trav.fp        = fp;

		/*	Do recursive traversal.	*/
		traverse(
			trie->root,
			&trav,
			0);

		/*	Do check. . .	*/
		if (trav.count != trie->total)
			fprintf(
				stderr,
				"%s %d : Error - trie internal error (total != count).\n",
				__FILE__,
				__LINE__);

		/*	Free buffer.	*/
		if ( trav.bfr )
			free( trav.bfr );
	}
	fputs("}", fp);

	/*	Return no error.	*/
	return 0;
}

static
void
destroy(
	PyObject	*self)
{
	auto		TrieObject		*trie = (TrieObject *) self;

	/*	Free the trie.	*/
	if ( trie->root )
		freeTrie( trie->root );
}

/*=============================================================================
||	Python Sequence Protocol Methods.
=============================================================================*/

static
int
trie_sq_length(
	PyObject	*self)
{
	return ((TrieObject *) self)->total;
}

static
void
doConcat(
	char		*str,
	PyObject	*obj,
	PyObject	*data)
{
	auto	TrieObject	*trie = (TrieObject *) data;
	auto	unsigned	count;

	/*	Insert into other.	*/
	trie->root = insert(
					trie->root,
					str,
					&trie->total,
					&count,
					obj);
}

static
PyObject	*
trie_sq_concat(
	PyObject	*self,
	PyObject	*other)
{
	register	int				i;
	auto		TrieObject		*trie = (TrieObject *) self;
	auto		TrieObject		*tail = (TrieObject *) other;
	auto		Traversal		trav;
	auto		TrieObject		*newTrie;

	/*	If other is not a trie, return error.	*/
	if (! is_TrieObject( tail ))
		TrieError( "Concatenate object is not a trie" )

	/*	Copy to new.	*/
	newTrie = (TrieObject *) Trie(NULL, NULL);
	newTrie->root  = copy( trie->root );
	newTrie->total = trie->total;

	/*	Build traversal control block.	*/
	trav.operation = TRIE_INTERNAL;
	trav.bfr       = NULL;
	trav.len       = 0;
	trav.count     = 0;
	trav.data      = (GenObject) newTrie;
	trav.doFunc    = doConcat;

	/*	Do recursive traversal.	*/
	traverse(
		tail->root,
		&trav,
		0);

	/*	Free buffer.	*/
	if ( trav.bfr )
		free( trav.bfr );

	/*	Increment reference count, return object.	*/
	return (PyObject *) newTrie;
}

static
void
doSequence(
	TrieNode		*node,
	SeqTraversal	*trav,
	int				count,
	int				depth)
{
	register	int		i;
	auto		int		tmp;

	/*	Loop through current node.	*/
	for (i = 0; i < node->used; i++)
	{
		/*	Fix up buffer.	*/
		if (depth >= trav->len)
		{
			trav->len += 16;
			trav->bfr  = realloc(trav->bfr, trav->len);
		}

		/*	Put character in buffer.	*/
		trav->bfr[depth]     = node->els[i].c;
		trav->bfr[depth + 1] = '\0';

		/*	Keep track of count.	*/
		tmp = count + node->els[i].subCount;

		/*	If we found the proper range, go for it.	*/
		if (trav->lo < tmp)
		{
			/*	If we are at the proper count, then we found it.	*/
			if ( node->els[i].data )
			{
				/*	If this is the proper data element, return it.	*/
				if (count == trav->lo)
				{
					PyObject	*ret;

					/*	Build and return tuple.	*/
					switch ( trav->operation )
					{
					case TRIE_ASS_SLICE:
						/*	Decrement reference count.	*/
						if ( node->els[i].data )
							Dispose( node->els[i].data );

						/*	Pop front of list.	*/
						ret = PyList_GetItem(trav->list, trav->count++);

						/*	Store new data.	*/
						Attach( ret );
						node->els[i].data = ret;

						/*	Increment, look for completion of range.	*/
						if (++trav->lo >= trav->hi)
							return;
						break;
					case TRIE_SLICE:
						/*	Build the tuple.	*/
						ret = Py_BuildValue(
							"(sO)", trav->bfr,
							node->els[i].data);

						/*	Append tuple to list.	*/
						ListAppend(trav->list, ret);

						/*	Increment, look for completion of range.	*/
						if (++trav->lo >= trav->hi)
							return;
						break;
					case TRIE_ASS_ITEM:
						/*	Decrement reference count.	*/
						if ( node->els[i].data )
							Dispose( node->els[i].data );

						/*	Store new data.	*/
						Attach( trav->list );
						node->els[i].data = trav->list;
						break;
					case TRIE_ITEM:
						ret = Py_BuildValue(
							"(sO)", trav->bfr,
							node->els[i].data);
						trav->list = ret;
						return;
					}
				}

				/*	There is a data element, but this one isn't
				*	what I'm looking for.
				*/
				count++;
			}

			/*	Recurse.	*/
			if ( node->els[i].child )
			{
				/*	Recurse.	*/
				doSequence(
					node->els[i].child,
					trav,
					count,
					depth + 1);
			}
		}

		/*	Save count of what has gone before.	*/
		count = tmp;
	}
}

static
PyObject	*
trie_sq_item(
	PyObject	*self,
	int			lo)
{
	auto		GenObject		ret;
	auto		SeqTraversal	trav;
	auto		TrieObject		*trie = (TrieObject *) self;

	/*	If index out of range, return None.	*/
	if (lo < 0 || lo >= trie->total)
		TrieError( "Trie error, index out of range" )

	/*	Get setup for sequence traversal.	*/
	trav.operation = TRIE_ITEM;
	trav.bfr       = NULL;
	trav.len       = 0;
	trav.lo        = lo;
	trav.hi        = lo + 1;
	trav.list      = NULL;

	/*	Get sequence item.	*/
	doSequence(trie->root, &trav, 0, 0);

	/*	Free buffer.	*/
	if ( trav.bfr )
		free( trav.bfr );
	return trav.list;
}

static
PyObject	*
trie_sq_slice(
	PyObject	*self,
	int			lo,
	int			hi)
{
	auto		SeqTraversal	trav;
	auto		TrieObject		*trie = (TrieObject *) self;

	/*	Condition inputs.	*/
	lo = (lo < 0) ? 0 : lo;
	lo = (lo > trie->total) ? trie->total : lo;
	hi = (hi < 0) ? 0 : hi;
	hi = (hi > trie->total) ? trie->total : hi;

	/*	Get setup for sequence traversal.	*/
	trav.operation = TRIE_SLICE;
	trav.bfr       = NULL;
	trav.len       = 0;
	trav.list      = PyList_New( 0 );
	trav.lo        = lo;
	trav.hi        = hi;

	/*	If index out of range, return empty list.	*/
	if (lo > hi)
		return trav.list;

	/*	Get slice.	*/
	doSequence(trie->root, &trav, 0, 0);

	/*	Free buffer.	*/
	if ( trav.bfr )
		free( trav.bfr );
	return trav.list;
}

static
int
trie_sq_ass_item(
	PyObject	*self,
	int			lo,
	PyObject	*data)
{
	auto		GenObject		ret;
	auto		SeqTraversal	trav;
	auto		TrieObject		*trie = (TrieObject *) self;

	/*	If index out of range, return None.	*/
	if (lo < 0 || lo >= trie->total)
	{
		SetError( "Trie item assignment error." );
		return -1;
	}

	/*	Get setup for sequence traversal.	*/
	trav.operation = TRIE_ASS_ITEM;
	trav.bfr       = NULL;
	trav.len       = 0;
	trav.lo        = lo;
	trav.hi        = lo + 1;
	trav.list      = data;

	/*	Remap by sequence number.	*/
	doSequence(trie->root, &trav, 0, 0);

	/*	Free buffer.	*/
	if ( trav.bfr )
		free( trav.bfr );
	return 0;
}

static
int
trie_sq_ass_slice(
	PyObject	*self,
	int			lo,
	int			hi,
	PyObject	*data)
{
	auto		SeqTraversal	trav;
	auto		TrieObject		*trie = (TrieObject *) self;

	/*	Check to make sure enough data exists to do the assignment.	*/
	if (! PyList_Check( data ) || PyList_Size( data ) != (hi - lo))
	{
		SetError( "Trie slice assignment error" );
		return -1;
	}

	/*	Condition inputs.	*/
	lo = (lo < 0) ? 0 : lo;
	lo = (lo > trie->total) ? trie->total : lo;
	hi = (hi < 0) ? 0 : hi;
	hi = (hi > trie->total) ? trie->total : hi;

	/*	Get setup for sequence traversal.	*/
	trav.operation = TRIE_ASS_SLICE;
	trav.bfr       = NULL;
	trav.len       = 0;
	trav.count     = 0;
	trav.list      = data;
	trav.lo        = lo;
	trav.hi        = hi;

	/*	If index out of range, return None.	*/
	if (lo > hi)
		return 0;

	/*	Assign slice.	*/
	doSequence(trie->root, &trav, 0, 0);

	/*	Free buffer.	*/
	if ( trav.bfr )
		free( trav.bfr );
	return 0;
}

/*=============================================================================
||	Python Mapping Protocol Methods.
=============================================================================*/

static
int
trie_mp_length(
	PyObject	*self)
{
	return ((TrieObject *) self)->total;
}

static
PyObject	*
trie_mp_subscript(
	PyObject	*self,
	PyObject	*subScript)
{
	auto	char		*cp;
	auto	PyObject	*retObj;
	auto	TrieObject	*trie = (TrieObject *) self;

	/*	If the subscript is an integer object,
	*	call the sequence subscript function.
	*/
	if ( PyInt_Check( subScript ) )
		return trie_sq_item(self, (int) PyInt_AsLong( subScript ));

	/*	Check for proper type (Python String Object) of subscript.	*/
	if (! PyString_Check( subScript ))
		TrieError( "Trie subscript must be a string object." )

	/*	Now get a 'C' string out of the Python String Object.	*/
	cp = PyString_AsString( subScript );

	/*	Find the string in the trie.	*/
	if (trie->root == NULL || (retObj = find(trie->root, cp)) == NULL)
		TrieError( "Trie Subscript Error" );

	/*	Increment reference count, return object.	*/
	Py_INCREF( retObj );
	return retObj;
}

static
int
trie_mp_ass_subscript(
	PyObject	*self,
	PyObject	*subScript,
	PyObject	*data)
{
	auto	char		*cp;
	auto	int			count;
	auto	GenObject	obj;
	auto	TrieObject	*trie = (TrieObject *) self;

	/*	If the subscript is an integer object,
	*	call the sequence subscript function.
	*/
	if ( PyInt_Check( subScript ) )
		return trie_sq_ass_item(self, (int) PyInt_AsLong( subScript ), data);

	/*	Check for proper type (Python String Object) of subscript.	*/
	if (! PyString_Check( subScript ))
	{
		SetError( "Trie Subscript must be integer or string." );
		return -1;
	}

	/*	Now get a 'C' string out of the Python String Object.	*/
	cp = PyString_AsString( subScript );

	/*	Do insert-or-remap.	*/
	if (data == NULL || data == Py_None)
	{
		/*	Remove the node.	*/
		count = 0;
		obj = removeNode(
				trie->root,
				cp,
				&count,
				&trie->total);
	}
	else
		trie->root = insert(trie->root, cp, &trie->total, &count, data);

	/*	Increment reference count, return object.	*/
	return 0;
}

PySequenceMethods	TrieObjectSequenceMethods = {
	(inquiry)			trie_sq_length,
	(binaryfunc)		trie_sq_concat,
	(ssizeargfunc) 		0,				/*	sq_repeat.	*/
	(ssizeargfunc) 		trie_sq_item,
	(intintargfunc) 	trie_sq_slice,
	(intobjargproc) 	trie_sq_ass_item,
	(intintobjargproc)	trie_sq_ass_slice
};

PyMappingMethods	TrieObjectMappingMethods = {
	(inquiry)			trie_mp_length,
	(binaryfunc)		trie_mp_subscript,
	(objobjargproc)		trie_mp_ass_subscript
};

PyTypeObject		TrieObjectType = {
/*  Type Header.    */
    PyObject_HEAD_INIT( NULL )
    0,
    "Trie",
    sizeof( TrieObject ),
    0,

/*  Standard methods.   */
    (destructor)  destroy,
    (printfunc)   print,
    (getattrfunc) getattr,
    (setattrfunc) 0,
    (cmpfunc)     0,
    (reprfunc)    0,

/*  Sequence methods.   */
    0,
    &TrieObjectSequenceMethods,
    &TrieObjectMappingMethods,

/*  Extras. */
    0,
	0,
	0,
	0,
	0,
};

static  struct  PyMethodDef module_methods[] = {
	{	"Trie",		Trie,		1	},
	{	NULL,		NULL,		0	}
};

void
inittrie()
{
    auto    PyObject    *mod;
    auto    PyObject    *dict;

	/*	Set type.	*/
	TrieObjectType.ob_type = &PyType_Type;

    /*  Create module object and get module dictionary object.  */
    mod  = Py_InitModule("trie", module_methods);
    dict = PyModule_GetDict( mod );

    /*  Build error object. */
    trieErrorObject = PyErr_NewException("trie.error", NULL, NULL);
    PyDict_SetItemString(dict, "error", trieErrorObject);

    /*  Check for errors.   */
    if ( PyErr_Occurred())
        Py_FatalError("Cannot initialize trie module.");
}
